@@ -7,17 +7,16 @@ use crate::protocol::{
     self, read_auth, read_hello, Ack, ControlChannelCmd, DataChannelCmd, Hello, UdpTraffic,
     HASH_WIDTH_IN_BYTES,
 };
-use crate::transport::{TcpTransport, Transport};
+use crate::transport::{SocketOpts, TcpTransport, Transport};
 use anyhow::{anyhow, bail, Context, Result};
 use backoff::backoff::Backoff;
 use backoff::ExponentialBackoff;
 
 use rand::RngCore;
 use std::collections::HashMap;
-use std::net::SocketAddr;
 use std::sync::Arc;
 use std::time::Duration;
-use tokio::io::{self, copy_bidirectional, AsyncWriteExt};
+use tokio::io::{self, copy_bidirectional, AsyncReadExt, AsyncWriteExt};
 use tokio::net::{TcpListener, TcpStream, UdpSocket};
 use tokio::sync::{broadcast, mpsc, RwLock};
 use tokio::time;
@@ -34,6 +33,7 @@ type Nonce = protocol::Digest; // Also called `session_key`
 const TCP_POOL_SIZE: usize = 8; // The number of cached connections for TCP servies
 const UDP_POOL_SIZE: usize = 2; // The number of cached connections for UDP services
 const CHAN_SIZE: usize = 2048; // The capacity of various chans
+const HANDSHAKE_TIMEOUT: u64 = 5; // Timeout for transport handshake
 
 // The entrypoint of running a server
 pub async fn run_server(
@@ -111,7 +111,7 @@ impl<'a, T: 'static + Transport> Server<'a, T> {
             config,
             services: Arc::new(RwLock::new(generate_service_hashmap(config))),
             control_channels: Arc::new(RwLock::new(ControlChannelMap::new())),
-            transport: Arc::new(T::new(&config.transport).await?),
+            transport: Arc::new(T::new(&config.transport)?),
         })
     }
 
@@ -162,15 +162,28 @@ impl<'a, T: 'static + Transport> Server<'a, T> {
                         }
                         Ok((conn, addr)) => {
                             backoff.reset();
-                            debug!("Incomming connection from {}", addr);
 
-                            let services = self.services.clone();
-                            let control_channels = self.control_channels.clone();
-                            tokio::spawn(async move {
-                                if let Err(err) = handle_connection(conn, addr, services, control_channels).await.with_context(||"Failed to handle a connection to `server.bind_addr`") {
-                                    error!("{:?}", err);
+                            // Do transport handshake with a timeout
+                            match time::timeout(Duration::from_secs(HANDSHAKE_TIMEOUT), self.transport.handshake(conn)).await {
+                                Ok(conn) => {
+                                    match conn.with_context(|| "Failed to do transport handshake") {
+                                        Ok(conn) => {
+                                            let services = self.services.clone();
+                                            let control_channels = self.control_channels.clone();
+                                            tokio::spawn(async move {
+                                                if let Err(err) = handle_connection(conn, services, control_channels).await {
+                                                    error!("{:?}", err);
+                                                }
+                                            }.instrument(info_span!("handle_connection", %addr)));
+                                        }, Err(e) => {
+                                            error!("{:?}", e);
+                                        }
+                                    }
+                                },
+                                Err(e) => {
+                                    error!("Transport handshake timeout: {}", e);
                                 }
-                            }.instrument(info_span!("handle_connection", %addr)));
+                            }
                         }
                     }
                 },
@@ -187,6 +200,8 @@ impl<'a, T: 'static + Transport> Server<'a, T> {
             }
         }
 
+        info!("Shutdown");
+
         Ok(())
     }
 
@@ -215,7 +230,6 @@ impl<'a, T: 'static + Transport> Server<'a, T> {
 // Handle connections to `server.bind_addr`
 async fn handle_connection<T: 'static + Transport>(
     mut conn: T::Stream,
-    addr: SocketAddr,
     services: Arc<RwLock<HashMap<ServiceDigest, ServerServiceConfig>>>,
     control_channels: Arc<RwLock<ControlChannelMap<T>>>,
 ) -> Result<()> {
@@ -223,8 +237,7 @@ async fn handle_connection<T: 'static + Transport>(
     let hello = read_hello(&mut conn).await?;
     match hello {
         ControlChannelHello(_, service_digest) => {
-            do_control_channel_handshake(conn, addr, services, control_channels, service_digest)
-                .await?;
+            do_control_channel_handshake(conn, services, control_channels, service_digest).await?;
         }
         DataChannelHello(_, nonce) => {
             do_data_channel_handshake(conn, control_channels, nonce).await?;
@@ -235,12 +248,13 @@ async fn handle_connection<T: 'static + Transport>(
 
 async fn do_control_channel_handshake<T: 'static + Transport>(
     mut conn: T::Stream,
-    addr: SocketAddr,
     services: Arc<RwLock<HashMap<ServiceDigest, ServerServiceConfig>>>,
     control_channels: Arc<RwLock<ControlChannelMap<T>>>,
     service_digest: ServiceDigest,
 ) -> Result<()> {
-    info!("New control channel incomming from {}", addr);
+    info!("Try to handshake a control channel");
+
+    T::hint(&conn, SocketOpts::for_control_channel());
 
     // Generate a nonce
     let mut nonce = vec![0u8; HASH_WIDTH_IN_BYTES];
@@ -248,22 +262,24 @@ async fn do_control_channel_handshake<T: 'static + Transport>(
 
     // Send hello
     let hello_send = Hello::ControlChannelHello(
-        protocol::CURRENT_PROTO_VRESION,
+        protocol::CURRENT_PROTO_VERSION,
         nonce.clone().try_into().unwrap(),
     );
     conn.write_all(&bincode::serialize(&hello_send).unwrap())
         .await?;
+    conn.flush().await?;
 
     // Lookup the service
-    let services_guard = services.read().await;
-    let service_config = match services_guard.get(&service_digest) {
+    let service_config = match services.read().await.get(&service_digest) {
         Some(v) => v,
         None => {
             conn.write_all(&bincode::serialize(&Ack::ServiceNotExist).unwrap())
                 .await?;
             bail!("No such a service {}", hex::encode(&service_digest));
         }
-    };
+    }
+    .to_owned();
+
     let service_name = &service_config.name;
 
     // Calculate the checksum
@@ -285,10 +301,6 @@ async fn do_control_channel_handshake<T: 'static + Transport>(
         );
         bail!("Service {} failed the authentication", service_name);
     } else {
-        let service_config = service_config.clone();
-        // Drop the rwlock as soon as possible when we're done with it
-        drop(services_guard);
-
         let mut h = control_channels.write().await;
 
         // If there's already a control channel for the service, then drop the old one.
@@ -297,14 +309,15 @@ async fn do_control_channel_handshake<T: 'static + Transport>(
         // the client to reconnect.
         if h.remove1(&service_digest).is_some() {
             warn!(
-                "Dropping previous control channel for digest {}",
-                hex::encode(service_digest)
+                "Dropping previous control channel for service {}",
+                service_name
             );
         }
 
         // Send ack
         conn.write_all(&bincode::serialize(&Ack::Ok).unwrap())
             .await?;
+        conn.flush().await?;
 
         info!(service = %service_config.name, "Control channel established");
         let handle = ControlChannelHandle::new(conn, service_config);
@@ -321,12 +334,20 @@ async fn do_data_channel_handshake<T: 'static + Transport>(
     control_channels: Arc<RwLock<ControlChannelMap<T>>>,
     nonce: Nonce,
 ) -> Result<()> {
+    debug!("Try to handshake a data channel");
+
     // Validate
     let control_channels_guard = control_channels.read().await;
     match control_channels_guard.get2(&nonce) {
         Some(handle) => {
+            T::hint(&conn, SocketOpts::from_server_cfg(&handle.service));
+
             // Send the data channel to the corresponding control channel
-            handle.data_ch_tx.send(conn).await?;
+            handle
+                .data_ch_tx
+                .send(conn)
+                .await
+                .with_context(|| "Data channel for a stale control channel")?;
         }
         None => {
             warn!("Data channel has incorrect nonce");
@@ -339,6 +360,7 @@ pub struct ControlChannelHandle<T: Transport> {
     // Shutdown the control channel by dropping it
     _shutdown_tx: broadcast::Sender<bool>,
     data_ch_tx: mpsc::Sender<T::Stream>,
+    service: ServerServiceConfig,
 }
 
 impl<T> ControlChannelHandle<T>
@@ -358,27 +380,6 @@ where
         // Store data channel creation requests
         let (data_ch_req_tx, data_ch_req_rx) = mpsc::unbounded_channel();
 
-        match service.service_type {
-            ServiceType::Tcp => tokio::spawn(
-                run_tcp_connection_pool::<T>(
-                    service.bind_addr.clone(),
-                    data_ch_rx,
-                    data_ch_req_tx.clone(),
-                    shutdown_tx.subscribe(),
-                )
-                .instrument(Span::current()),
-            ),
-            ServiceType::Udp => tokio::spawn(
-                run_udp_connection_pool::<T>(
-                    service.bind_addr.clone(),
-                    data_ch_rx,
-                    data_ch_req_tx.clone(),
-                    shutdown_tx.subscribe(),
-                )
-                .instrument(Span::current()),
-            ),
-        };
-
         // Cache some data channels for later use
         let pool_size = match service.service_type {
             ServiceType::Tcp => TCP_POOL_SIZE,
@@ -391,11 +392,48 @@ where
             };
         }
 
+        let shutdown_rx_clone = shutdown_tx.subscribe();
+        let bind_addr = service.bind_addr.clone();
+        match service.service_type {
+            ServiceType::Tcp => tokio::spawn(
+                async move {
+                    if let Err(e) = run_tcp_connection_pool::<T>(
+                        bind_addr,
+                        data_ch_rx,
+                        data_ch_req_tx,
+                        shutdown_rx_clone,
+                    )
+                    .await
+                    .with_context(|| "Failed to run TCP connection pool")
+                    {
+                        error!("{:?}", e);
+                    }
+                }
+                .instrument(Span::current()),
+            ),
+            ServiceType::Udp => tokio::spawn(
+                async move {
+                    if let Err(e) = run_udp_connection_pool::<T>(
+                        bind_addr,
+                        data_ch_rx,
+                        data_ch_req_tx,
+                        shutdown_rx_clone,
+                    )
+                    .await
+                    .with_context(|| "Failed to run TCP connection pool")
+                    {
+                        error!("{:?}", e);
+                    }
+                }
+                .instrument(Span::current()),
+            ),
+        };
+
         // Create the control channel
         let ch = ControlChannel::<T> {
             conn,
             shutdown_rx,
-            service,
+            service: service.clone(),
             data_ch_req_rx,
         };
 
@@ -412,6 +450,7 @@ where
         ControlChannelHandle {
             _shutdown_tx: shutdown_tx,
             data_ch_tx,
+            service,
         }
     }
 }
@@ -436,7 +475,11 @@ impl<T: Transport> ControlChannel<T> {
                 val = self.data_ch_req_rx.recv() => {
                     match val {
                         Some(_) => {
-                            if let Err(e) = self.conn.write_all(&cmd).await.with_context(||"Failed to write data cmds") {
+                            if let Err(e) = self.conn.write_all(&cmd).await.with_context(||"Failed to write control cmds") {
+                                error!("{:?}", e);
+                                break;
+                            }
+                            if let Err(e) = self.conn.flush().await.with_context(|| "Failed to flush control cmds") {
                                 error!("{:?}", e);
                                 break;
                             }
@@ -453,7 +496,7 @@ impl<T: Transport> ControlChannel<T> {
             }
         }
 
-        info!("Control channel shuting down");
+        info!("Control channel shutdown");
 
         Ok(())
     }
@@ -467,8 +510,11 @@ fn tcp_listen_and_send(
     let (tx, rx) = mpsc::channel(CHAN_SIZE);
 
     tokio::spawn(async move {
-        let l = backoff::future::retry(listen_backoff(), || async {
+        // FIXME: Respect shutdown signal
+        let l = backoff::future::retry_notify(listen_backoff(), || async {
             Ok(TcpListener::bind(&addr).await?)
+        }, |e, duration| {
+            error!("{:?}. Retry in {:?}", e, duration);
         })
         .await
         .with_context(|| "Failed to listen for the service");
@@ -509,10 +555,9 @@ fn tcp_listen_and_send(
                         }
                         Ok((incoming, addr)) => {
                             // For every visitor, request to create a data channel
-                            if let Err(e) = data_ch_req_tx.send(true).with_context(|| "Failed to send data chan create request") {
+                            if data_ch_req_tx.send(true).with_context(|| "Failed to send data chan create request").is_err() {
                                 // An error indicates the control channel is broken
                                 // So break the loop
-                                error!("{:?}", e);
                                 break;
                             }
 
@@ -530,6 +575,8 @@ fn tcp_listen_and_send(
                 }
             }
         }
+
+        info!("TCPListener shutdown");
     }.instrument(Span::current()));
 
     rx
@@ -555,6 +602,8 @@ async fn run_tcp_connection_pool<T: Transport>(
             break;
         }
     }
+
+    info!("Shutdown");
     Ok(())
 }
 
@@ -567,9 +616,18 @@ async fn run_udp_connection_pool<T: Transport>(
 ) -> Result<()> {
     // TODO: Load balance
 
-    let l: UdpSocket = backoff::future::retry(listen_backoff(), || async {
-        Ok(UdpSocket::bind(&bind_addr).await?)
-    })
+    // FIXME: Respect shutdown signal
+    let l: UdpSocket = backoff::future::retry_notify(
+        listen_backoff(),
+        || async {
+            Ok(UdpSocket::bind(&bind_addr)
+                .await
+                .with_context(|| "Failed to listen for the service")?)
+        },
+        |e, duration| {
+            warn!("{:?}. Retry in {:?}", e, duration);
+        },
+    )
     .await
     .with_context(|| "Failed to listen for the service")?;
 
@@ -594,10 +652,10 @@ async fn run_udp_connection_pool<T: Transport>(
             },
 
             // Forward outbound traffic from the client to the visitor
-            t = UdpTraffic::read(&mut conn) => {
-                let t = t?;
+            hdr_len = conn.read_u8() => {
+                let t = UdpTraffic::read(&mut conn, hdr_len?).await?;
                 l.send_to(&t.data, t.from).await?;
-            },
+            }
 
             _ = shutdown_rx.recv() => {
                 break;
@@ -605,5 +663,7 @@ async fn run_udp_connection_pool<T: Transport>(
         }
     }
 
+    debug!("UDP pool dropped");
+
     Ok(())
 }
